"""
Module that defines shared Exceptions between blocks.
Should be used to raise specific `sys.exit` codes as below:

![Exceptions](assets/exceptions.png)

Make use of the decorator in your `run` method to provide appropriate
standardized exit codes:

```python
@catch_exceptions(logger)
def run():
    do_stuff()
```
"""
import functools
import sys
from enum import Enum

from logging import Logger
from .logging import get_logger


class SupportedErrors(Enum):
    """
    Enumerator of standardized error codes for the UP42 platform.

    - `INPUT_PARAMETERS_ERROR` (2)
        - User provided wrong or inconsistent configuration parameters.
    - `NO_INPUT_ERROR` (3)
        - Block did not find input data, e.g. no data in the
            requested area (data block), no features in data.json (processing block).
    - `WRONG_INPUT_ERROR` (4)
        - Input data is unsuitable, e.g. a processing block
            expects 16 bit but receives 8 bit.
    - `API_CONNECTION_ERROR` (5)
        - API which is used by the block is down or changed its interface.
    - `NO_OUTPUT_ERROR` (6)
        - After applying all processing step no data results
            to be provided to the user.
    - `ERR_INCORRECT_ERRCODE` (100)
    """

    INPUT_PARAMETERS_ERROR = (
        2
        # User provided wrong or inconsistent configuration parameters.
    )
    NO_INPUT_ERROR = (
        3
        # Block did not find input data, e.g. no data in the
        # requested area (data block), no features in data.json (processing block).
    )
    WRONG_INPUT_ERROR = (
        4
        # Input data is unsuitable, e.g. a processing block
        # expects 16 bit but receives 8 bit.
    )
    API_CONNECTION_ERROR = (
        5
        # API which is used by the block is down or changed its interface.
    )
    NO_OUTPUT_ERROR = (
        6
        # After applying all processing step no data results
        # to be provided to the user.
    )
    ERR_INCORRECT_ERRCODE = 100  # No error code defined.


class UP42Error(Exception):
    """
    The UP42 error base class.
    """

    def __init__(self, error_code: SupportedErrors, message=""):
        """
        Example:
            ```python
            if not results:
                raise UP42Error(SupportedErrors.NO_OUTPUT_ERROR)
            ```

        Arguments:
            error_code: A SupportedErrors instance, i.e.
                `SupportedErrors.NO_OUTPUT_ERROR`
            message: An optional message to be logged.
        """
        # Raise a separate exception in case the error code passed isn't specified in the ErrorCodes enum
        if not isinstance(error_code, SupportedErrors):
            msg = "Error code passed in the error_code param must be of type {0}"
            raise UP42Error(SupportedErrors.ERR_INCORRECT_ERRCODE, message)

        # Storing the error code on the exception object
        self.error_code = error_code

        # storing the traceback which provides useful information about where the exception occurred
        self.traceback = sys.exc_info()

        # Prefixing the error code to the exception message
        msg = "[{0}] {1}".format(error_code.name, message)

        Exception.__init__(self, msg)


# Disable unused function argument due to wrapper
# pylint: disable=unused-argument
def catch_exceptions(logger: Logger = get_logger(__name__)):
    """
    A decorator that wraps the passed in function and
    returns the appropriate exit code.

    Arguments:
        logger: A logging instance. Can be generated with
            `blockutils.logging.get_logger(__name__)`
    """

    def decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except UP42Error as e:
                if e.error_code == SupportedErrors.INPUT_PARAMETERS_ERROR:
                    log_exception(function, logger, "Incorrect input parameters.")
                    sys.exit(SupportedErrors.INPUT_PARAMETERS_ERROR.value)
                elif e.error_code == SupportedErrors.NO_INPUT_ERROR:
                    log_exception(function, logger, "No input data found.")
                    sys.exit(SupportedErrors.NO_INPUT_ERROR.value)
                elif e.error_code == SupportedErrors.WRONG_INPUT_ERROR:
                    log_exception(function, logger, "Unsuitable input data found.")
                    sys.exit(SupportedErrors.WRONG_INPUT_ERROR.value)
                elif e.error_code == SupportedErrors.API_CONNECTION_ERROR:
                    log_exception(
                        function,
                        logger,
                        "Connection error with API. Please try again later.",
                    )
                    sys.exit(SupportedErrors.API_CONNECTION_ERROR.value)
                elif e.error_code == SupportedErrors.NO_OUTPUT_ERROR:
                    log_exception(function, logger, "No output data created.")
                    sys.exit(SupportedErrors.NO_OUTPUT_ERROR.value)
                else:
                    log_exception(function, logger, "Undefined error code.")
                    sys.exit(SupportedErrors.ERR_INCORRECT_ERRCODE.value)
            except MemoryError:
                log_exception(
                    function,
                    logger,
                    "Memory limit surpassed. Please make sure to check the docs for block limitations.",
                )
                sys.exit(137)
            except Exception:  # pylint: disable=broad-except
                log_exception(function, logger)
                sys.exit(1)

        def log_exception(function, logger, err="Generic exception."):
            logger.exception(err)

        return wrapper

    return decorator



"""
Common methods shared between blocks, especially useful for directory handling
and block parameter/query input.
"""
import json
import os
from typing import Callable, Union, List
from contextlib import ContextDecorator
from pathlib import Path
from enum import Enum
import shutil
import base64
from datetime import datetime, timedelta

import rasterio as rio
from rasterio import warp
from shapely.geometry import box, mapping
from dateutil.parser import parse
from geojson import FeatureCollection, Feature

from .stac import STACQuery
from .logging import get_logger

logger = get_logger(__name__)


def ensure_data_directories_exist():
    """Creates required directories for any block input and output (`/tmp/input`,
    `tmp/output`, `/tmp/quicklooks`).
    """
    Path("/tmp/input/").mkdir(parents=True, exist_ok=True)
    Path("/tmp/output/").mkdir(parents=True, exist_ok=True)
    Path("/tmp/quicklooks/").mkdir(parents=True, exist_ok=True)


def setup_test_directories(
    test_dir: Union[str, Path], clean_subdirs: bool = True
) -> List[Path]:
    """Creates given test directory and empty subdirs "input", "output", "quicklooks".

    Args:
        test_dir: A directory to store temporary files (usually `/tmp` or `/tmp/e2e_test`)
        clean_subdirs: Remove all files in the subdirs "input", "output", "quicklooks".
    Returns:
        List of sub directories Paths depending onthe selection in sub_dirs.
    """
    test_dir = Path(test_dir)
    test_dir.mkdir(parents=True, exist_ok=True)

    subdirs = ["input", "output", "quicklooks"]
    subdir_paths = [test_dir / subdir for subdir in subdirs]
    for subdir_path in subdir_paths:
        if clean_subdirs:
            # Should not remove other files/dirs in test_dir, since often /tmp.
            subdir_path.mkdir(parents=True, exist_ok=True)
            try:
                shutil.rmtree(subdir_path)
            # Deleting subfolder sometimes does not work in temp, then remove all subfiles.
            except (PermissionError, OSError):
                files_to_delete = subdir_path.rglob("*.*")
                for file_path in files_to_delete:
                    file_path.unlink()
        subdir_path.mkdir(parents=True, exist_ok=True)

    return subdir_paths


class TestDirectoryContext(ContextDecorator):
    """Yields the test directory making sure folders exist and cleans up when
    context is closed.
    """

    def __init__(self, test_dir: Path):
        """
        Example:
            ```python
            with TestDirectoryContext(Path("/tmp")) as test_dir:
                block.process(dir=Path("/tmp"))
            ```

        Arguments:
            test_dir: A directory to store temporary files (usually `/tmp`
                or `/tmp/e2e_test`)
        """
        self.test_dir = test_dir

    def __enter__(self) -> Path:
        """Context entry point.

        Returns:
            Temporary test directory.
        """
        setup_test_directories(self.test_dir)
        return self.test_dir

    def __exit__(self, *exc):
        """Context exit point. Cleans up test subdirs."""
        setup_test_directories(test_dir=self.test_dir)
        return False


def load_query(validator: Callable = lambda x: True) -> STACQuery:
    """Get the query for the current task directly from the task parameters
    in `UP42_TASK_PARAMETERS` environment variable.

    Example:
        ```python
        def val(data: dict) -> bool:
            # Ensure bbox is defined.
            return "bbox" in data

        query = load_query(val)
        ```

    Arguments:
        validator: Callable that returns if the loaded query is valid.

    Returns:
        A `STACQuery` object initialized with the parameters if valid.
    """
    data: str = os.environ.get("UP42_TASK_PARAMETERS", "{}")
    logger.debug(f"Raw task parameters from UP42_TASK_PARAMETERS are: {data}")
    query_data = json.loads(data)
    return STACQuery.from_dict(query_data, validator)


def load_params() -> dict:
    """Get the parameters for the current task directly from the task
    parameters parameters in `UP42_TASK_PARAMETERS` environment variable.

    Returns:
        Dictionary of task parameters.
    """
    data: str = os.environ.get("UP42_TASK_PARAMETERS", "{}")
    logger.debug(f"Fetching parameters for this block: {data}")
    if data == "":
        data = "{}"
    return json.loads(data)


def load_metadata() -> FeatureCollection:
    """Get the geojson metadata input.

    Returns:
        Object defining input features for block.
    """
    ensure_data_directories_exist()
    if os.path.exists("/tmp/input/data.json"):
        with open("/tmp/input/data.json") as fp:
            data = json.loads(fp.read())

        features = []
        for feature in data["features"]:
            features.append(Feature(**feature))

        return FeatureCollection(features)
    else:
        return FeatureCollection([])


def save_metadata(result: FeatureCollection):
    """Save the geojson metadata output.

    Arguments:
        result: Output feature collection.
    """
    ensure_data_directories_exist()
    with open("/tmp/output/data.json", "w") as fp:
        fp.write(json.dumps(result))


def update_extents(feat_coll: FeatureCollection) -> FeatureCollection:
    """
    Updates all geometry extents to reflect actual images

    Arguments:
        feat_coll: geojson Feature Collection

    Returns: A FeatureCollection where image extents reflect actual images
    """
    for feature in feat_coll.features:
        with rio.open(
            os.path.join("/tmp/output", feature.properties["up42.data_path"])
        ) as img_file:
            img_bounds = img_file.bounds
        bounds_trans = warp.transform_bounds(
            img_file.crs, {"init": "epsg:4326"}, *img_bounds
        )

        geom = box(*bounds_trans)
        feature["geometry"] = mapping(geom)
        feature["bbox"] = geom.bounds

    return feat_coll


class BlockModes(Enum):
    """
    Types of block modes: DRY_RUN or DEFAULT.

    Important:
        Find out more about job modes/block modes in our
        [documentation](https://docs.up42.com/reference/block-envvars.html#up42-job-mode).
    """

    DRY_RUN = "DRY_RUN"
    DEFAULT = "DEFAULT"


def get_block_mode() -> str:
    """Gets the task mode from environment variables. If no task mode is set,
    DEFAULT mode will be returned.

    Important:
        Find out more about job modes/block modes in our
        [documentation](https://docs.up42.com/reference/block-envvars.html#up42-job-mode).

    Returns:
        Block mode.
    """
    value = os.environ.get("UP42_JOB_MODE", BlockModes.DEFAULT.value)
    if value not in [mode.value for mode in BlockModes]:
        value = "DEFAULT"
    return value


def get_block_info() -> dict:
    """Gets the Block Info variable as a dictionary.

    Returns:
        Block info as a dict.
    """
    value_str = str(os.environ.get("UP42_BLOCK_INFO"))
    value_dict = json.loads(value_str)

    return value_dict


def encode_str_base64(string: str) -> str:
    """
    A function that encodes strings in base64. The primary use case is passing complex environment variables into
    docker containers. In cases where these env variables are complex json objects including a number of different
    special characters the process of getting them unharmed into a docker container sometimes fails. Encoding them
    in base64 is a save method to solve this problem.

    Arguments:
        string: A, potentially complex, non-encoded string
    Returns:
        A base64-encoded string
    """
    return base64.b64encode(string.encode("ascii")).decode("ascii")


def decode_str_base64(string: str) -> str:
    """
    Inverter function for encode_str_base64

    Arguments:
        string: A base64-encoded string
    Returns:
        A decoded string
    """

    str_encoded_byte = bytes(string, "utf-8")
    str_decoded_byte = base64.b64decode(str_encoded_byte)
    return str_decoded_byte.decode("utf-8")


def get_timeperiod(
    duration: int = 365,
    start: int = -365,
    start_date: str = None,
    end_date: str = None,
) -> str:
    """
    Generates time period string from any combinations of duration(days) & start(days),
    duration(days) & start_date or start_date & end_date parameters. Defaults to 1 year till today.

    Most relevant for selecting timeperiods for testing of rolling archives.

    Args:
        duration: Time period duration in days.
        start: Time period start from today in days.
        start_date: Time period start date, format string YYYY-MM-DD.
        end_date: Time period end date, format string YYYY-MM-DD.

    Returns:
        time period string with variable time and fixed clock time, in the format
        "2019-07-01T00:00:00+00:00/2029-07-10T23:59:59+00:00"
    """
    if start_date is not None and end_date is not None:
        start_date = parse(start_date).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        end_date = parse(end_date).replace(hour=0, minute=0, second=0, microsecond=0)
    elif start_date is not None and duration is not None:
        start_date = parse(start_date).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        end_date = start_date + timedelta(days=+duration)
    elif duration is not None and start is not None:
        today = datetime.today().replace(hour=0, minute=0, second=0, microsecond=0)
        start_date = today + timedelta(days=+start)
        end_date = start_date + timedelta(days=+duration)
    else:
        raise ValueError(
            "Only the combinations `1. duration(months) & start(months), "
            "2. duration(months) & start_date or 3. start_date & end_date` are allowed."
        )

    start_date = start_date.strftime("%Y-%m-%d")
    end_date = end_date.strftime("%Y-%m-%d")

    time = f"{start_date}T00:00:00+00:00/{end_date}T23:59:59+00:00"
    return time



"""
Important:
    DEPRECATED in favour of `blockutils.blocks.DataBlock`.
    See documentation [here](blocks.md#blockutils.blocks.DataBlock).

Abstract classes for data blocks.
"""
from .stac import STACQuery


class AbstractFetcher:
    """
    Base abstract fetcher.
    """

    def fetch(self, query: STACQuery, dry_run: bool = False):
        raise NotImplementedError


class AbstractAOIClippedFetcher(AbstractFetcher):
    """
    AOIClipped abstract fetcher.
    """

    def fetch(self, query: STACQuery, dry_run: bool = False):
        raise NotImplementedError


class AbstractFullSceneFetcher(AbstractFetcher):
    """
    FullScene abstract fetcher.
    """

    def fetch(self, query: STACQuery, dry_run: bool = False):
        raise NotImplementedError



"""
Utilities to handle exotic file formats (not GeoTiff), particularly DIMAP and NETCDF.
"""
# pylint: disable=C0411
from typing import Tuple
from pathlib import Path
import glob

import rasterio as rio
from rasterio.windows import Window
import numpy as np
from geojson import Feature

from blockutils.exceptions import UP42Error, SupportedErrors
from blockutils.datapath import set_data_path


def update_jsonfile(feature: Feature, output_name: str) -> Feature:
    """
    This method updates properties of a feature with a new file path.

    Arguments:
        feature: Input feature object.
        output_name: Output file path from main working folder (i.e. output.tif).

    Returns:
        Feature object with the updated data path to output_name.
    """
    new_feature = Feature(geometry=feature["geometry"], bbox=feature["bbox"])
    prop_dict = feature["properties"]
    meta_dict = {
        k: v
        for k, v in prop_dict.items()
        if not (k.startswith("up42.") or k.startswith("custom."))
    }
    new_feature["properties"] = meta_dict
    set_data_path(new_feature, output_name)
    return new_feature


class DimapFile:
    """
    Base class for handling DIMAP files.
    """

    def __init__(self, base_path: Path = Path("/tmp/input")):
        """
        Arguments:
            base_path: Main input working folder.
        """
        self.base_path = base_path

    def check_dtype(self, feature: Feature) -> str:
        """
        This method opens the xml file related to the image and check for the
        data type of the image.
        Args:
            feature: Input feature.

        Returns:
            Dtype of the image.

        """
        dimap_file_id = feature.properties.get("up42.data_path")
        dimap_path = self.base_path.joinpath(dimap_file_id)

        img_files = glob.glob(str(dimap_path) + "/**/IMG_*", recursive=True)
        dimap_dirs = list(Path(img_files[0]).glob("DIM_*"))
        img_data_name = Path(img_files[0]).joinpath(dimap_dirs[0].name)

        accepted_dytpe = ["uint16", "uint8"]
        with rio.open(img_data_name) as src:
            img_dtype = src.profile["dtype"]

        if img_dtype not in accepted_dytpe:
            raise UP42Error(
                SupportedErrors.WRONG_INPUT_ERROR,
                f"Input dtype must be one of {accepted_dytpe}.",
            )

        return img_dtype

    @staticmethod
    def open_xml_file_with_rasterio(path: Path) -> rio.io.DatasetReader:
        """
        This method open xml file with rasterio.
        Args:
            path: Path to the xml file.

        Returns:
            Rasterio DatasetReader.
        """
        img_name = Path(path).joinpath(list(path.glob("DIM_*"))[0].name)

        return rio.open(img_name)

    def dimap_file_path(self, feature: Feature) -> Tuple[Path, Path]:
        """
        This methods returns the folder name of the MS and PAN files.
        (e.g. IMG_PHR1B_MS_001, IMG_PHR1B_PS_002)

        Arguments:
            feature: Input feature.

        Returns:
            Path to multispectral and panchromatic directory.
        """
        dimap_file_id = feature.properties.get("up42.data_path")
        dimap_path = self.base_path.joinpath(dimap_file_id)

        ms_files = glob.glob(str(dimap_path) + "/**/IMG_*_MS_*", recursive=True)
        ms_file = Path(ms_files[0]) or Path()
        pan_files = glob.glob(str(dimap_path) + "/**/IMG_*_P_*", recursive=True)
        pan_file = Path(pan_files[0]) or Path()

        return ms_file, pan_file

    def dimap_8bit_file_path(self, feature: Feature) -> Path:
        """
        This methods returns the folder name of the PMS file.
        (e.g. IMG_PHR1B_PMS_001)

        Arguments:
            feature: Input feature.

        Returns:
            Path to pan-sharpened multispectral image directory.
        """
        dimap_file_id = feature.properties.get("up42.data_path")
        dimap_path = self.base_path.joinpath(dimap_file_id)

        pms_files = glob.glob(str(dimap_path) + "/**/IMG_*_PMS_*", recursive=True)
        pms_file = Path(pms_files[0]) or Path()

        return pms_file

    def get_meta_input(self, feature: Feature, mode: str) -> Tuple[dict, dict]:
        """
        This method returns the profile of the input image.

        Arguments:
            feature: Input feature object.
            mode: `ms` or `pan` depending on the profile to be returned.

        Returns:
            Rasterio profile object.
        """
        if mode in ["ms", "pan"]:
            ms_data_path, pan_data_path = self.dimap_file_path(feature)
            if mode == "ms":
                data_src = self.open_xml_file_with_rasterio(ms_data_path)
            if mode == "pan":
                data_src = self.open_xml_file_with_rasterio(pan_data_path)
        elif mode == "pms":
            pms_data_path = self.dimap_8bit_file_path(feature)
            data_src = self.open_xml_file_with_rasterio(pms_data_path)

        return data_src.profile

    def get_dim_xml_path(self, feature: Feature) -> Tuple:
        """
        This method returns the path to the input image.
        (e.g. IMG_PHR1B_PMS_001/DIM_PHR1B_MS_201810161039434_ORT_15007a44-dffa-41fe-c109-0d4fecabd40b-001.XM)
        Arguments:
            feature: Input feature object.

        Returns:
            Paths to multispectral and panchromatic XML files.
        """
        ms_path, pan_path = self.dimap_file_path(feature)
        ms_name = Path(ms_path).joinpath(list(ms_path.glob("DIM_*"))[0].name)
        pan_name = Path(pan_path).joinpath(list(pan_path.glob("DIM_*"))[0].name)

        return ms_name, pan_name

    def read_image_as_raster(self, feature: Feature, mode: str) -> np.ndarray:
        """
        This method returns the input image in a numpy array format.

        Arguments:
            feature: Input feature object.
            mode: `ms` or `pan` depending on the array to be returned.

        Returns:
            Imagery in numpy array format.
        """
        if mode in ["ms", "pan"]:
            ms_data_path, pan_data_path = self.dimap_file_path(feature)
            if mode == "ms":
                data_src = self.open_xml_file_with_rasterio(ms_data_path)
            if mode == "pan":
                data_src = self.open_xml_file_with_rasterio(pan_data_path)
        elif mode == "pms":
            pms_data_path = self.dimap_8bit_file_path(feature)
            data_src = self.open_xml_file_with_rasterio(pms_data_path)

        return data_src.read()

    # pylint: disable=too-many-locals
    def clip_and_read_image_as_raster(self, feature: Feature, mode: str, bounds):
        """
        This method returns the input image in a numpy array format.
        """
        if mode in ["ms", "pan"]:
            ms_data_path, pan_data_path = self.dimap_file_path(feature)
            if mode == "ms":
                data_src = self.open_xml_file_with_rasterio(ms_data_path)
            if mode == "pan":
                data_src = self.open_xml_file_with_rasterio(pan_data_path)
        elif mode == "pms":
            pms_data_path = self.dimap_8bit_file_path(feature)
            data_src = self.open_xml_file_with_rasterio(pms_data_path)

        data_src_transform = data_src.transform
        lonmin, latmin, lonmax, latmax = bounds
        rows, cols = rio.transform.rowcol(
            data_src_transform, (lonmin, lonmax), (latmin, latmax)
        )
        clipped_data = data_src.read(
            window=Window(
                col_off=min(cols),
                row_off=min(rows),
                width=max(cols) - min(cols),
                height=max(rows) - min(rows),
            )
        )
        return (
            clipped_data,
            min(cols),
            min(rows),
            max(cols) - min(cols),
            max(rows) - min(rows),
        )


class NetCDFFile:
    """
    Base class for handling NETCDF files.
    """

    def __init__(self, base_path: Path = Path("/tmp/input")):
        """
        Arguments:
            base_path: Main input working folder.
        """
        self.base_path = base_path

    def path_to_nc_file(self, feature: Feature) -> Path:
        """
        This methods returns the file name based on the given params.

        Arguments:
            feature: Input feature object.

        Returns:
            Path to input NETCDF file.
        """
        nc_file_id = feature.properties.get("up42.data_path")
        nc_path = self.base_path.joinpath(nc_file_id)

        return nc_path



"""
Utility class to generate synthetic images, especially usefull for testing purposes.
"""

import uuid
from pathlib import Path
from typing import Union, Optional

import rasterio
from rasterio import crs
from rasterio.transform import from_origin
import skimage.draw
from scipy import signal
import numpy as np

from .logging import get_logger

logger = get_logger(__name__)

# Define standard land cover classes and their standard stats wrt radiance for a 4-band product
LC_CLASSES = {
    "water": {"avg": [170, 300, 450, 150], "std": [10, 10, 10, 10]},
    "bare_ground": {"avg": [600, 600, 600, 900], "std": [100, 100, 100, 100]},
    "built_up": {"avg": [800, 800, 800, 1000], "std": [150, 150, 150, 150]},
    "forest": {"avg": [250, 450, 450, 1300], "std": [30, 30, 30, 30]},
    "non-forest_vegetation": {
        "avg": [300, 500, 500, 1100],
        "std": [100, 100, 100, 100],
    },
}


class SyntheticImage:
    # pylint: disable=E1137
    # pylint: disable=R0913
    # pylint: disable=too-many-instance-attributes
    """
    Create synthetic GeoTIFF test image. An image created this way cannot recreate all characteristics of a
    real geospatial image, but if cleverly created can avoid having to use golden files for a
    long list of cases.
    """

    def __init__(
        self,
        xsize: int,
        ysize: int,
        num_bands: int,
        data_type: str,
        out_dir: Path = Path("."),
        coord_ref_sys: int = 3857,
        nodata: Union[int, float] = 0,
        nodata_fill: int = 0,
    ):
        """
        Arguments:
            xsize: Number of pixels in x-direction.
            ysize: Number of pixels in y-direction.
            num_bands: Number of image bands.
            data_type: Rasterio datatype as string.
            out_dir: Path where the image should be created (default ".").
            coord_ref_sys: EPSG identifier of used coordinate reference system (default 3837).
            nodata: Value representing nodata within each raster band, default is 0. If set to -1 no nodata value set.
            nodata_fill: number of no data pixels to set in top left image (in x and y).

        Example:
            ```python
            input_img, an_array = SyntheticImage(
                10, 10, 4, "uint16", input_dir, nodata_fill=3
            ).create(seed=45)
            ```
        """

        self.xsize = xsize
        self.ysize = ysize
        self.num_bands = num_bands
        self.data_type = data_type
        self.out_dir = out_dir
        self.crs = coord_ref_sys
        self.nodata_fill = nodata_fill
        if nodata == -1:
            self.nodata = None
        else:
            self.nodata = nodata

    def add_img_pattern(self, seed: Optional[int]) -> np.ndarray:
        """
        Simulate a five classes optical image.

        Arguments:
            seed: A random seed number. Ensures reproducibility.

        Returns:
            Numpy array representing simulated image.
        """
        if seed is not None:
            np.random.seed(seed)

        image, _ = skimage.draw.random_shapes(
            (self.ysize, self.xsize),
            max_shapes=50,
            min_shapes=25,
            multichannel=False,
            allow_overlap=True,
            random_seed=seed,
        )
        # Assign shape values to output classes
        image[image < 55] = 1
        image[(image >= 55) & (image < 105)] = 2
        image[(image >= 105) & (image < 155)] = 3
        image[(image >= 155) & (image < 205)] = 4
        image[(image >= 205) & (image <= 255)] = 5

        # Create bands having relevant values for all output classes
        bands = []
        band_idx = 0
        while band_idx < self.num_bands:
            data_ar = np.zeros_like(image, dtype=self.data_type)

            for class_idx, lc_class in enumerate(LC_CLASSES.values(), 1):
                # Add Gaussian noise
                try:
                    mask_ar = np.random.normal(
                        lc_class["avg"][band_idx],
                        lc_class["std"][band_idx],
                        image.shape,
                    )
                except IndexError:
                    mask_ar = np.random.normal(
                        lc_class["avg"][-1], lc_class["std"][-1], image.shape
                    )
                data_ar[image == class_idx] = mask_ar[
                    image == class_idx
                ]  # pylint: disable=unsubscriptable-object
            # Apply median filter to simulate spatial autocorrelation
            data_ar = (signal.medfilt(data_ar)).astype(self.data_type)
            data_ar = np.clip(data_ar, 1, None)
            bands.append(data_ar)
            band_idx += 1

        return bands

    def create(
        self,
        seed: Optional[int] = None,
        transform: rasterio.Affine = from_origin(1470996, 6914001, 2.0, 2.0),
        file_name: str = None,
        band_desc: list = None,
    ) -> tuple:
        """
        Creates a synthethic image file with a given seed. Returns a tuple with
        (path to file, array).
        Transform is set by default to `from_origin(1470996, 6914001, 2.0, 2.0)`. Pass
        another Affine transform if needed.

        Arguments:
            seed: A random seed number. Ensures reproducibility.
            transform: An Affine transform for the image to be generated.
            file_name: A name for the created file.
            band_desc: List with descriptions (strings) for each band.

        Returns:
            Path to the output image, numpy array of image values.
        """
        band_list = self.add_img_pattern(seed)

        if not file_name:
            filepath = self.out_dir.joinpath(str(uuid.uuid4()) + ".tif")
        else:
            filepath = self.out_dir.joinpath(file_name + ".tif")

        # Even though we don't want a nodata value to be set, we still need a value that represents it for image
        # creation purposes
        if self.nodata is None:
            nodata_val = 0
        else:
            nodata_val = self.nodata  # type: ignore
        with rasterio.open(
            filepath,
            "w",
            driver="GTiff",
            height=self.ysize,
            width=self.xsize,
            count=self.num_bands,
            dtype=str(band_list[0].dtype),
            crs=crs.CRS.from_epsg(self.crs),
            transform=transform,
            nodata=self.nodata,
        ) as out_img:
            for band_id, layer in enumerate(band_list):
                layer[0 : self.nodata_fill, 0 : self.nodata_fill] = nodata_val
                out_img.write_band(band_id + 1, layer)
                if band_desc is not None:
                    try:
                        out_img.set_band_description(band_id + 1, band_desc[band_id])
                    except IndexError:
                        logger.debug(
                            "Number of band descriptions does not match number of bands"
                        )

        return filepath, np.array(band_list)



"""
Utility class to handle windowed read and write rasterio operations.
"""

from typing import Union, Iterable

import rasterio as rio
from rasterio.windows import Window
from rasterio.enums import Resampling
import numpy as np


class WindowsUtil:
    """
    Utility class to handle raster IO in windows. Can do regular windows, buffered
    windows and transform windows.
    """

    def __init__(self, rio_ds: Union[rio.io.DatasetReader, rio.io.DatasetWriter]):
        """
        Initialize the instance with a rasterio dataset (read or write).

        Arguments:
            rio_ds: An open rasterio dataset.
        """
        self.rio_ds = rio_ds
        self.windows = rio_ds.block_windows(1)

    def windows_regular(self) -> Iterable[Window]:
        """
        Method that returns regular rasterio windows.

        Returns:
            An iterable of rasterio windows.
        """
        for _, window in self.windows:
            yield window

    def windows_buffered(self, buffer: int = 0) -> Iterable[Window]:
        """
        Method that returns buffered windows with a given int buffer.

        Arguments:
            buffer: A buffer size in pixels.

        Returns:
            An iterable of rasterio windows.
        """
        for _, window in self.windows:
            buffered_window = self.buffer_window(window, buffer)
            yield window, buffered_window

    def windows_transformed(
        self, dst_transform: rio.Affine, dst_px_height: int, dst_px_width: int
    ) -> Iterable[Window]:
        """
        Method that returns windows with a transformed into a given rasterio Dataset
        Affine transform.

        Arguments:
            dst_transform: A target transform to use to generate windows.
            dst_px_height: The destination dataset pixel height.
            dst_px_width: The destination dataset pixel width.

        Returns:
            An iterable of rasterio windows.
        """
        for _, window in self.windows:
            transformed_window = self.transform_window(window, dst_transform)
            transformed_window = self.limit_window_to_raster_bounds(
                transformed_window, dst_px_height, dst_px_width
            )
            yield window, transformed_window

    def upsample_window_array(
        self,
        low_res_window: Window,
        high_res_window: Window,
        resampling_method: Resampling = Resampling.bilinear,
    ) -> np.array:
        """
        Method that returns an upsampled array to a given window, with an adapatable
        resampling method.

        Arguments:
            low_res_window: The low resolution window.
            high_res_window: The upsampled high resolution target window.
            resampling_method: The resampling method to use.

        Returns:
            An upsampled window numpy array.
        """
        up_ar = self.rio_ds.read(
            window=low_res_window,
            out_shape=(
                self.rio_ds.count,
                high_res_window.height,
                high_res_window.width,
            ),
            resampling=resampling_method,
        )
        return up_ar

    def buffer_window(self, window: Window, buffer: int) -> Window:
        """
        Buffers a window with a set number of pixels (buffer) in every possible direction
        given a shape of the overall raster file where window is derived from.

        For instance, if window matches the shape of the source raster file, the shape
        of the source raster file is returned in window format.

        Arguments:
            window: Original window (for instance from block).
            buffer: Number of pixels to buffer by where possible.

        Returns:
            A buffered window.
        """
        row_slice, col_slice = window.toslices()

        can_row_start = row_slice.start - buffer
        can_row_stop = row_slice.stop + buffer

        can_col_start = col_slice.start - buffer
        can_col_stop = col_slice.stop + buffer

        out_row_slice = slice(can_row_start, can_row_stop)
        out_col_slice = slice(can_col_start, can_col_stop)

        return self.limit_window_to_raster_bounds(
            Window.from_slices(out_row_slice, out_col_slice, boundless=True)
        )

    def transform_window(self, window: Window, out_transform: rio.Affine) -> Window:
        """
        Transforms a window using the geopgraphical bounds to generate a new window
        given the Affine transform.

        Arguments:
            window: Original window.
            out_transform: Transformation of the destination dataset.

        Returns:
            The original window transformed into the destination dataset.
        """
        out_window = rio.windows.from_bounds(
            *rio.windows.bounds(window, transform=self.rio_ds.transform),
            transform=out_transform
        ).round_shape("ceil")
        return out_window

    def limit_window_to_raster_bounds(
        self, window: Window, dst_height: int = None, dst_width: int = None
    ) -> Window:
        """
        Make sure the window fits in the dst raster. If not "clips" window to the
        bounds of the raster.
        This method is required because when applying a transform in the Windows
        precision in the Affine transformation can cause inconsistencies on the
        size of the windows in relation to the final output file.

        Arguments:
            window: Original window.
            dst_height: Destination dataset height.
            dst_width: Destination dataset width.

        Returns:
            The original window limited to the destination dataset.
        """
        if dst_height is None:
            dst_height = int(self.rio_ds.height)
        if dst_width is None:
            dst_width = int(self.rio_ds.width)

        window_slices_row, window_slices_col = window.toslices()
        result_row = [int(window_slices_row.start), int(window_slices_row.stop)]
        result_col = [int(window_slices_col.start), int(window_slices_col.stop)]

        if window_slices_row.start < 0:
            result_row[0] = 0
        if window_slices_row.stop > dst_height:
            result_row[1] = dst_height

        if window_slices_col.start < 0:
            result_col[0] = 0
        if window_slices_col.stop > dst_width:
            result_col[1] = dst_width

        return Window.from_slices(result_row, result_col)

    def crop_array_to_window(
        self, buffered_array: np.array, window: Window, window_buffer: Window
    ) -> np.array:
        """
        Crops an array created with the windows_buffered to the extent of
        window. Makes use of the a higher res Affine to "reproject" the original
        window to the extent of window_buffer.

        Arguments:
            buffered_array: Buffered array read with window_buffer.
                Has same shape as buffered window.
            window: Original window.
                Buffered window cropped with a given buffer.
            window_buffer: Buffered window.

        Returns:
            A cropped numpy array.
        """
        buffer_transform = rio.windows.transform(window_buffer, self.rio_ds.transform)
        window_transformed = rio.windows.from_bounds(
            *rio.windows.bounds(window, transform=self.rio_ds.transform),
            transform=buffer_transform
        )

        # row, col
        (
            window_buffer_slices_row,
            window_buffer_slices_col,
        ) = window_transformed.toslices()

        slice_col_start = int(round(window_buffer_slices_col.start))
        slice_row_start = int(round(window_buffer_slices_row.start))

        slice_col_stop = int(round(window_buffer_slices_col.stop))
        slice_row_stop = int(round(window_buffer_slices_row.stop))

        cropped_array = buffered_array[
            :, slice_row_start:slice_row_stop, slice_col_start:slice_col_stop
        ]
        return cropped_array



"""
Helper for downloading, processing and merging tiles from tile based data sources.
"""
import io
import uuid
import time
from pathlib import Path
from typing import List, Callable
from contextlib import contextmanager
from dataclasses import dataclass, field

import concurrent.futures
import numpy as np
import rasterio
from rasterio import merge as riomerge
from rasterio.enums import ColorInterp
import mercantile
from mercantile import Tile
import requests


from .logging import get_logger
from .raster import to_cog, create_multiband_tif
from .exceptions import UP42Error, SupportedErrors

logger = get_logger(__name__)


class TileIsEmptyError(TypeError):
    pass


class TileNotFetchedError(TypeError):
    pass


@dataclass
class TileMergeHelperSettings:
    """
    This class wraps all defaultable attributes of the TileMergeHelper.
    """

    req_kwargs: dict = field(default_factory=dict)
    tile_size: int = field(default=256)
    crs: str = field(default="EPSG:4326")
    parallelize: bool = field(default=False)
    retries_for_no_fetched_tiles: int = field(default=3)
    max_threads: int = field(default=10)
    polling_cycle: int = field(default=5)


class TileMergeHelper(TileMergeHelperSettings):
    """
    This class is a helper class that allows to fetch tiles via a req function, process them via a process function
    (a custom process function can be called, or the default process function will be used) ,and finally merge them
    into a single image in COG format.
    """

    def __init__(
        self,
        tiles: List[Tile],
        req: Callable,
        process: Callable = None,
        **kwargs,
    ):
        self.tiles = tiles
        self.req = req
        self.process = process
        super().__init__(**kwargs)

    @staticmethod
    def _process(response: requests.Response, tile: Tile) -> Path:
        """
        This method is default method for processing a tile. It simply writes a tile which can be in
        different formats such as "png", "jpeg", or etc in the tif format to make sure all metadata and transform
        information are saved so that merging process can be performed.
        Args:
            response: response from API side after fetching a tile
            tile: A tuple of x,y,z information for the tile. For instance: Tile(x=41955, y=101467, z=18)
        """
        temp_tile: rasterio.MemoryFile = io.BytesIO(response.content)
        temp_tile_filename = Path(f"/tmp/output/{str(uuid.uuid4())}.tif")

        bands: list = []
        with rasterio.open(temp_tile) as image:
            for i in range(image.count):
                bands.append(image.read(i + 1))
            tile_meta = image.meta

        tile_transform = rasterio.transform.from_bounds(
            *mercantile.xy_bounds(tile),
            width=tile_meta.get("width"),
            height=tile_meta.get("height"),
        )

        tile_meta.update(driver="GTiff", crs="EPSG:3857", transform=tile_transform)
        if np.any(bands):
            with rasterio.open(temp_tile_filename, "w", **tile_meta) as output_tile:
                for idx, band in enumerate(bands):
                    output_tile.write(band, idx + 1)
        else:
            raise TileIsEmptyError

        return temp_tile_filename

    def tile_worker(
        self,
        tile: Tile,
        tile_tif_list: List[Path],
        valid_tiles: List[Tile],
        tiles_not_fetched: List[Tile],
    ):
        """
        This method fetches the tile from a specific url and applies processing on it.
        Final result should be a path to the tile written in tif format. If user did not specify a process
        method, the default _process method will called.
        Args:
            tile: A tuple of x,y,z information for the tile. For instance: Tile(x=41955, y=101467, z=18)
            tile_tif_list: List for saving the tiles that were fetched and stored in the tif format.
            valid_tiles: List of tiles that are NOT empty.
            tiles_not_fetched: List of tiles that were not fetched due to API issues.
        """
        try:
            response = self.req(tile, **self.req_kwargs)
            try:
                if self.process:
                    tile_tif_list.append(self.process(response, tile))
                else:
                    tile_tif_list.append(self._process(response, tile))
                valid_tiles.append(tile)
            except TileIsEmptyError:
                logger.info(f"{tile} is emtpy, Skipping ...")
        except TileNotFetchedError:
            tiles_not_fetched.append(tile)

    def loop_over_tiles(self):
        """
        This method iterates over a list of tiles and applies tile_worker method to fetch and process them.
        In the case of not being able to fetch the tile (API issues) a retrying mechanism will happen to
        retry fetching those tiles again. Also for speeding up the fetching process a parallelization option
        has been implemented. Moreover, if a tile is empty, it will be skipped.
        Returns:
            tile_tif_list: List of path to tiles which are written in tif format.
            valid_tiles: List of tiles that are NOT empty.
        """
        tile_tif_list = []
        valid_tiles = []
        tiles_not_fetched = []

        active_tiles = self.tiles.copy()

        retry_round = 0
        while retry_round <= self.retries_for_no_fetched_tiles:
            # It initially starts with active_tiles, and among them, there are some tiles that are not fetched
            # which will be saved to tiles_not_fetched list. Then when retrying started, the tile in this list
            # will be assigned to active_tiles (active_tiles will be rewritten) to do the whole fetching process again.
            if retry_round:
                if tiles_not_fetched:
                    time.sleep(self.polling_cycle)
                    active_tiles = tiles_not_fetched.copy()
                    tiles_not_fetched = []
                    logger.info(
                        f"Now starting {retry_round}. round of retrying failures"
                    )
                    logger.info(f"{len(active_tiles)}. tiles will be retried")
                else:
                    break

            if not self.parallelize:
                for tile in active_tiles:
                    self.tile_worker(
                        tile, tile_tif_list, valid_tiles, tiles_not_fetched
                    )
            else:
                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=self.max_threads
                ) as executor:
                    future_list = []
                    for tile in active_tiles:
                        logger.info(f"Tile {tile} added to executor...")
                        future_list.append(
                            executor.submit(
                                lambda t: self.tile_worker(*t),
                                [tile, tile_tif_list, valid_tiles, tiles_not_fetched],
                            )
                        )
                    for future in concurrent.futures.as_completed(future_list):
                        try:
                            future.result()
                        except Exception as e:
                            raise e
            retry_round += 1

        logger.info(
            f"There are {len(valid_tiles)} valid data tiles out of "
            f"{len(self.tiles)}"
        )
        if tiles_not_fetched:
            raise UP42Error(
                SupportedErrors.API_CONNECTION_ERROR,
                "Retrying was unsuccessful. OneAtlas API returned error code 500",
            )
        return tile_tif_list, valid_tiles

    @contextmanager
    def tile_dataset(self):

        tile_tif_list, valid_tiles = self.loop_over_tiles()

        try:
            yield tile_tif_list, valid_tiles
        finally:
            for tile_tif in tile_tif_list:
                tile_tif.unlink()

    @staticmethod
    def merge_tiles(tile_file_list):
        """
        This method merges the list of tif files and returns the merged array,
        transform and meta for the final writing.
        Args:
            tile_file_list: List of path to tif files. For instance: ["/a/b/1234.tif", "/a/c/5678.tif"]
        """
        if not tile_file_list:
            raise UP42Error(SupportedErrors.NO_INPUT_ERROR, "All tiles are empty.")

        tile_file_list_str = [str(tif_file) for tif_file in tile_file_list]
        merged_array, merged_transform = riomerge.merge(tile_file_list_str)
        with rasterio.open(tile_file_list[0]) as src:
            temp_meta = src.meta.copy()

        return merged_array, merged_transform, temp_meta

    def write_merged_tiles(
        self,
        output_path,
        merged_array,
        merged_transform,
        meta,
        tile_size,
        apply_to_cog: bool = True,
    ):
        """
        This method writes the merged tiles in a file. The final result will be in COG format.
        Args:
            output_path: Path to save the final image.
            merged_array: Numpy array of merged tiles.
            merged_transform: Transform for merged tiles.
            meta: Metadata for writing the final image.
            tile_size: Size of the tile. In most cases 256.
            apply_to_cog: Whether convert a tif to cog tif. Default is true.

        """
        _merged_shape: List = list(merged_array.shape)
        if _merged_shape[1] > tile_size and _merged_shape[2] > tile_size:
            while _merged_shape[1] % tile_size != 0:
                # X dimension not divisible by tile_size
                _merged_shape[1] -= 1
                # Remove one pixel
            while _merged_shape[2] % tile_size != 0:
                # Y dimension not divisible by tile_size
                _merged_shape[2] -= 1
                # Remove one pixel

        merged_shape = tuple(_merged_shape)

        meta.update(
            {
                "driver": "GTiff",
                "height": merged_shape[1],
                "width": merged_shape[2],
                "transform": merged_transform,
                "crs": self.crs,
            }
        )
        with rasterio.open(output_path, "w", **meta) as output:
            output.write(merged_array[:, : merged_shape[1], : merged_shape[2]])

        if apply_to_cog:
            to_cog(output_path)

    def get_final_image(self, output_path: Path, return_cog: bool = True):
        """The main method which merges list of tiles and produces the final merged image."""
        with self.tile_dataset() as (tile_tif_list, valid_tiles):
            merged_array, merged_transform, meta = self.merge_tiles(tile_tif_list)

        self.write_merged_tiles(
            output_path,
            merged_array,
            merged_transform,
            meta,
            self.tile_size,
            return_cog,
        )
        return valid_tiles


class MultiTileMergeHelper:
    """
    This class is a helper class that allows to pass list of TileMergeHelper object. This is mostly relevant
    when the datasets has multiple layers (e.g. SENTINEL1GRD with two polarization). Via this object, each layer will
    be handled separately (fetching tiles, processing them, and merging them) with an additional final step which
    appends this layers to one tif file to a create a single multi-bands tif. Also, the end result will be in COG
    format.
    """

    def __init__(self, tilemergehelper_list: List[TileMergeHelper]):
        self.tilemergehelper_list = tilemergehelper_list

    def get_multiband_tif(
        self,
        filename_path: Path,
        band_descriptions: List[str] = None,
        drop_nodata: bool = False,
        colorinterp: List[ColorInterp] = None,
        return_cog=True,
    ):
        outputs_path = []
        valid_tiles = {}
        for i, tilemergehelper in enumerate(self.tilemergehelper_list):
            tmp = f"/tmp/output/{str(uuid.uuid4())}.tif"
            valid_tiles[i] = tilemergehelper.get_final_image(
                Path(tmp), return_cog=False
            )
            outputs_path.append(Path(tmp))

        if band_descriptions:
            for old_key, new_key in zip(valid_tiles.keys(), band_descriptions):
                valid_tiles[new_key] = valid_tiles.pop(old_key)

        create_multiband_tif(
            outputs_path,
            filename_path,
            band_descriptions,
            drop_nodata,
            colorinterp,
            return_cog=return_cog,
        )
        for tmp_path in outputs_path:
            tmp_path.unlink()

        return valid_tiles

    @classmethod
    def from_req_kwargs(
        cls,
        tiles: List[Tile],
        req: Callable,
        kwargs_list: List[dict],
        process: Callable = None,
    ):
        tilemergehelper_list = []
        for kwargs in kwargs_list:
            tilemergehelper_list.append(TileMergeHelper(tiles, req, process, **kwargs))

        return cls(tilemergehelper_list)



"""
Base classes for building data and processing blocks. Includes basic functionality
including [exception handling](exceptions.md).

Example:
    ```python
    class AProcessingBlock(ProcessingBlock):
        def __init__(self, a_var):
            super().__init__()
            self.a_var = a_var

        def process(
            self, input_fc: FeatureCollection, process_dir: Path = Path("/tmp")
        ):
            self.a_var = 2
            return input_fc
    ```
    This class takes the input feature collection and return the same feature
    collection unchanged. The entrypoint would always be the `run` method.
"""

from abc import ABC, abstractmethod
from geojson import FeatureCollection, Feature

from .common import (
    load_query,
    load_params,
    load_metadata,
    save_metadata,
    ensure_data_directories_exist,
    BlockModes,
    get_block_mode,
)
from .exceptions import catch_exceptions

from .geometry import check_validity
from .stac import STACQuery


class DataBlock(ABC):
    """
    Base class for data blocks.
    """

    @abstractmethod
    def fetch(self, query: STACQuery, dry_run: bool = False) -> FeatureCollection:
        """
        Main worker method, for the given STAC query, fetch the results and save
        to output.

        Important:
            This method is an [abstract method](https://docs.python.org/3/library/abc.html)
            which means you are required to implement it in any inherited class.
        """
        raise NotImplementedError

    @classmethod
    @catch_exceptions()
    def run(cls, **kwargs):
        """
        This method is the main entry point for the data block.

        Allows for arbitrary kwargs that will be passed directy to the
        data block instance.

        Example:
            ```python3
            a_data_block.run(an_argument=15)
            ```
        """
        ensure_data_directories_exist()
        query: STACQuery = load_query()
        # check for geometry validity
        if query.bbox or query.intersects or query.contains:
            check_validity(query_geom=query.geometry())
        dry_run: bool = get_block_mode() == BlockModes.DRY_RUN.value
        result: FeatureCollection = cls(**kwargs).fetch(query=query, dry_run=dry_run)
        save_metadata(result)


class ProcessingBlock(ABC):
    """
    Base class for processing blocks.
    """

    @abstractmethod
    def process(self, input_fc: FeatureCollection) -> FeatureCollection:
        """
        Main worker method, for each input feature, process it and save to output.

        Important:
            This method is an [abstract method](https://docs.python.org/3/library/abc.html)
            which means you are required to implement it in any inherited class.
        """
        raise NotImplementedError

    @staticmethod
    def get_metadata(feature: Feature) -> dict:
        """
        Extracts metadata elements that need to be propagated to the output tif.
        """
        prop_dict = feature["properties"]
        meta_dict = {
            k: v
            for k, v in prop_dict.items()
            if not (k.startswith("up42.") or k.startswith("custom."))
        }
        return meta_dict

    @classmethod
    def from_dict(cls, kwargs):
        """
        Instantiate a class with a dictionary of parameters
        """
        return cls(**kwargs)

    @classmethod
    @catch_exceptions()
    def run(cls):
        """
        This method is the main entry point for the processing block.
        """
        ensure_data_directories_exist()
        params: dict = load_params()
        input_metadata: FeatureCollection = load_metadata()
        processing_block = cls.from_dict(params)
        result: FeatureCollection = processing_block.process(input_metadata)
        save_metadata(result)



"""
Common raster handling methods shared between blocks
"""
from pathlib import Path
from typing import List

import numpy as np
import rasterio as rio
from rasterio.io import DatasetReader
from rasterio.enums import ColorInterp
from rio_cogeo.cogeo import cog_translate
from rio_cogeo.profiles import cog_profiles

from blockutils.logging import get_logger

logger = get_logger(__name__)


def is_empty(path_to_image: Path, nodataval=0) -> bool:
    """
    Tests if a created geotiff image only consists of nodata or NaN values
    Converts NaN to nodata values as a side effect
    Args:
        path_to_image: Path object pointing to geotiff image
        nodataval: no data value, default is 0

    Returns: True if image is empty, False otherwise
    """
    with rio.open(str(path_to_image)) as img_file:
        data = img_file.read()
        np.nan_to_num(data, nan=nodataval, copy=False)
        return not np.any(data - nodataval)


def to_cog(path_to_image: Path, profile: str = "deflate", **options) -> bool:
    """
    Converts a GeoTIFF into a Cloud-optimized GeoTIFF
    :param path_to_image: path to GeoTIFF
    :param profile: compression profile
    :param options: additional kwargs
    :return: True if all went well
    """
    logger.info("Now converting to COG")
    tmp_file_path = Path(str(path_to_image) + ".tmp")
    path_to_image.rename(tmp_file_path)

    # Format creation option (see gdalwarp `-co` option)
    output_profile = cog_profiles.get(profile)
    output_profile.update(dict(BIGTIFF="IF_SAFER"))

    # Dataset Open option (see gdalwarp `-oo` option)
    config = dict(
        GDAL_NUM_THREADS="ALL_CPUS",
        GDAL_TIFF_INTERNAL_MASK=True,
        GDAL_TIFF_OVR_BLOCKSIZE="128",
    )

    cog_translate(
        str(tmp_file_path),
        str(path_to_image),
        output_profile,
        config=config,
        in_memory=False,
        quiet=False,
        **options,
    )
    tmp_file_path.unlink()
    return True


def _get_profile(
    list_tif_files_dataset: List[DatasetReader],
    band_descriptions: List[str] = None,
    drop_nodata: bool = False,
    colorinterp: List[ColorInterp] = None,
):
    count = 0
    for tif_dataset in list_tif_files_dataset:
        count += tif_dataset.count

    if band_descriptions:
        assert count == len(band_descriptions)

    if colorinterp:
        assert count == len(colorinterp)

    raster_profile = list_tif_files_dataset[0].profile
    raster_profile.update(count=count)
    if drop_nodata:
        raster_profile.pop("nodata")

    return raster_profile


def create_multiband_tif(
    list_tif_files: List[Path],
    filename_path: Path,
    band_descriptions: List[str] = None,
    drop_nodata: bool = False,
    colorinterp: List[ColorInterp] = None,
    return_cog: bool = True,
):
    """
    This method combines list of tif files and save them
    in one tif with each of tif files to be a separate band
    Args:
        list_tif_files: List of separate tif files.
        filename_path: Final path for a single multi-bands tif file.
        band_descriptions: Description of each band.
        drop_nodata: Ignore nodata in final output.
        use_colorinterp: Add color interpretation for each band.
        return_cog: Whether convert a tif to cog tif. Default is true.
    """
    with rio.Env():

        list_tif_files_dataset = [rio.open(tif_file) for tif_file in list_tif_files]

        raster_profile = _get_profile(
            list_tif_files_dataset, band_descriptions, drop_nodata, colorinterp
        )

        with rio.open(filename_path, "w", **raster_profile) as dst:
            for _, window in list_tif_files_dataset[0].block_windows(1):
                src = []
                for tif_dataset in list_tif_files_dataset:
                    src.append(tif_dataset.read(window=window))
                dst.write(np.concatenate(src, axis=0), window=window)

            if band_descriptions:
                for b_id, layer in enumerate(band_descriptions):
                    dst.set_band_description(b_id + 1, layer)
            if colorinterp:
                dst.colorinterp = colorinterp
    if return_cog:
        to_cog(filename_path)



import os
from pathlib import Path
import setuptools

PARENT_DIR = Path(__file__).resolve().parent


def set_directory():
    # CD to this directory, to simplify package finding
    abspath = os.path.abspath(__file__)
    dname = os.path.dirname(abspath)
    os.chdir(dname)


set_directory()

with open("docs/index.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="up42-blockutils",
    version=PARENT_DIR.joinpath("blockutils/_version.txt").read_text(encoding="utf-8"),
    author="UP42",
    author_email="support@up42.com",
    description="Block development toolkit for UP42",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://www.up42.com",
    packages=setuptools.find_packages(exclude=("tests", "apiutils")),
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
    ],
    install_requires=[
        "geojson",
        "numpy",
        "rasterio",
        "shapely",
        "area",
        "ciso8601",
        "scipy",
        "scikit-image",
        "rio-cogeo",
    ],
    python_requires=">=3.6, <3.9",
)
